@@ -1,10 +1,18 @@
 # -*- coding: utf-8 -*-
 from __future__ import absolute_import
 import os.path as osp
-import logging, subprocess, os, glob, shutil, time, datetime, argparse
+import logging, subprocess, os, glob, time, datetime, argparse
 from contextlib import contextmanager
 
+N_COPY_ATTEMPTS = 1
 DEFAULT_LOGGING_LEVEL = logging.WARNING
+N_SECONDS_SLEEP = 10
+
+INCLUDE_DIR = osp.join(osp.abspath(osp.dirname(__file__)), "include")
+
+def version():
+    with open(osp.join(INCLUDE_DIR, "VERSION"), "r") as f:
+        return(f.read().strip())
 
 def setup_logger(name='seutils'):
     if name in logging.Logger.manager.loggerDict:
@@ -56,69 +64,169 @@ def is_string(string):
         basestring = str
     return isinstance(string, basestring)
 
-N_SECONDS_SLEEP = 10
+ENV = None
+def set_env(env):
+    """
+    Sets the env in which command line arguments are ran by default
+    """
+    global ENV
+    ENV = env
 
-def run_command(cmd, env=None, dry=None, nonzero_exitcode_ok=False, n_retries=0, return_output_on_nonzero_exitcode=False):
+@contextmanager
+def env_context(env):
     """
-    Runs a command and captures output. Raises an exception on non-zero exit code,
-    except if nonzero_exitcode_ok is set to True.
+    Temporarily sets an environment, and then reverts to the old environment
+    """
+    global ENV
+    old_ENV = ENV
+    ENV = env
+    try:
+        yield None
+    finally:
+        ENV = old_ENV
+
+def add_env_kwarg(fn):
+    """
+    Function decorator that gives the function the `env` keyword argument
+    """
+    def wrapper(*args, **kwargs):
+        if 'env' in kwargs:
+            with env_context(kwargs.pop('env')):
+                return fn(*args, **kwargs)
+        else:
+            return fn(*args, **kwargs)
+    return wrapper
+
+RM_BLACKLIST = [
+    '/',
+    '/store',
+    '/store/user',
+    '/store/user/*',
+    ]
+RM_WHITELIST = []
+
+def rm_safety(fn):
+    """
+    Safety wrapper around any rm function: Raise an exception for some paths
+    """
+    import re
+    def wrapper(*args, **kwargs):
+        path = args[1]
+        if not has_protocol(path):
+            logger.error('Remote rm operation called on local path')
+            raise RmSafetyTrigger(path)
+        path = split_mgm(normpath(path))[1]
+        depth = path.count('/')
+        logger.debug('In rm_safety wrapper')
+        # Check if the passed `path` is in the blacklist:
+        for bl_path in RM_BLACKLIST:
+            if bl_path == path:
+                raise RmSafetyTrigger(path)
+            elif bl_path.count('/') != depth:
+                continue
+            elif re.match(bl_path.replace('*', '.*'), path):
+                raise RmSafetyTrigger(path)
+        # Check if the passed `path` is in the whitelist:
+        if RM_WHITELIST:
+            for wl_path in RM_WHITELIST:
+                if path.startswith(wl_path):
+                    break
+            else:
+                logger.error('Path is outside of the whitelist: ' + ', '.join(RM_WHITELIST))
+                raise RmSafetyTrigger(path)
+        return fn(*args, **kwargs)
+    return wrapper
+
+def listdir_check_isdir(fn):
+    """
+    Wrapper around listdir implementations, that first checks if the directory exists
+    """
+    def wrapper(*args, **kwargs):
+        if not kwargs.pop('assume_isdir', False):
+            if not args[0].isdir(args[1]):
+                raise Exception('Cannot listdir {0}: not a directory'.format(args[1]))
+        return fn(*args, **kwargs)
+    return wrapper
+
+def run_command_rcode_and_output(cmd, env=None, dry=None):
+    """Runs a command and captures output.
+    Returns return code and captured output.
     """
-    i_attempt = 0
     if dry is None: dry = DRYMODE
     if env is None: env = ENV
-    while True:
-        logger.info('%sIssuing command (attempt %s: %s)', '(dry) ' if dry else '', i_attempt, ' '.join(cmd))
-        if dry: return ''
-        process = subprocess.Popen(
-            cmd,
-            stdout=subprocess.PIPE,
-            stderr=subprocess.STDOUT,
-            env=env,
-            universal_newlines=True,
-            )
-        # Start running command and capturing output
-        output = []
-        for stdout_line in iter(process.stdout.readline, ''):
-            logger.debug('CMD: ' + stdout_line.strip('\n'))
-            output.append(stdout_line)
-        process.stdout.close()
-        process.wait()
-        returncode = process.returncode
-        # Return output only if command succeeded
-        if returncode == 0:
-            logger.info('Command exited with status 0 - all good')
-        else:
-            if nonzero_exitcode_ok:
-                logger.info('Command exited with status %s', returncode)
-                return output if return_output_on_nonzero_exitcode else returncode
-            else:
-                logger.error('Exit status {0} for command: {1}'.format(returncode, cmd))
-                logger.error('Output:\n%s', '\n'.join(output))
-                if i_attempt < n_retries:
-                    i_attempt += 1
-                    logger.error('Retrying attempt %s/%s in %s seconds...', i_attempt, n_retries, N_SECONDS_SLEEP)
-                    time.sleep(N_SECONDS_SLEEP)
-                    continue
-                else:
-                    raise subprocess.CalledProcessError(cmd, returncode)
-        return output
+    logger.info('%sIssuing command %s', '(dry) ' if dry else '', ' '.join(cmd))
+    if dry: return 0, '<dry output>'
+    process = subprocess.Popen(
+        cmd,
+        stdout=subprocess.PIPE,
+        stderr=subprocess.STDOUT,
+        env=env,
+        universal_newlines=True,
+        )
+    # Start running command and capturing output
+    output = []
+    for stdout_line in iter(process.stdout.readline, ''):
+        logger.debug('CMD: ' + stdout_line.strip('\n'))
+        output.append(stdout_line)
+    process.stdout.close()
+    process.wait()
+    return process.returncode, output
 
-def get_exitcode(cmd, env=None):
+def run_command_rcode_and_output_with_retries(cmd, *args, **kwargs):
+    """
+    Wrapper around run_command_rcode_and_output that repeats on a non-zero exit code
+    """
+    n_attempts = kwargs.pop('n_attempts', 1)
+    for i_attempt in range(n_attempts):
+        if n_attempts > 1:
+            logger.info(
+                'Running command %s with retries: attempt %s of %s',
+                cmd, i_attempt+1, n_attempts
+                )
+        rcode, output = run_command_rcode_and_output(cmd, *args, **kwargs)
+        if rcode != 0:
+            return rcode, output
+        if n_attempts > 1:
+            logger.info(
+                'Return code %s for attempt %s of %s',
+                rcode, i_attempt+1, n_attempts
+                )
+        if i_attempt+1 < n_attempts: time.sleep(N_SECONDS_SLEEP)
+    else:
+        if n_attempts > 1: logger.info('Non-zero return code after %s attempt(s)!', n_attempts)
+        return rcode, output
+
+def run_command(cmd, *args, **kwargs):
+    """
+    Main entrypoint for implementations.
+    Raises an exception on non-zero exit codes.
+
+    If `path` is specified as a keyword argument, it is used for a more descriptive
+    exception, but otherwise it is not used.
+    """
+    rcodes = kwargs.pop('rcodes', {})
+    path = kwargs.pop('path', '')
+    rcode, output = run_command_rcode_and_output_with_retries(cmd, *args, **kwargs)
+    if rcode == 0:
+        logger.info('Command exited with status 0 - all good')
+        return output
+    else:
+        logger.error(
+            '\033[31mExit status {0} for command {1}\nOutput:\n{2}\033[0m'
+            .format(rcode, cmd, '\n'.join(output))
+            )
+        if rcode in rcodes:
+            raise rcodes[rcode](path)
+        else:
+            raise NonZeroExitCode(rcode, cmd)
+
+def get_exitcode(cmd, *args, **kwargs):
     """
     Runs a command and returns the exit code.
     """
-    if env is None: env = ENV
-    if is_string(cmd): cmd = [cmd]
-    logger.debug('Getting exit code for "%s"', ' '.join(cmd))
-    if DRYMODE:
-        returncode = 0
-    else:
-        FNULL = open(os.devnull, 'w')
-        process = subprocess.Popen(cmd, stdout=FNULL, stderr=subprocess.STDOUT, env=env)
-        process.communicate()[0]
-        returncode = process.returncode
-    logger.debug('Got exit code %s', returncode)
-    return returncode
+    rcode, _ = run_command_rcode_and_output(cmd, *args, **kwargs)
+    logger.debug('Got exit code %s', rcode)
+    return rcode
 
 def bytes_to_human_readable(num, suffix='B'):
     """
@@ -275,8 +383,57 @@ def format(path, mgm=None):
     """
     if is_ssh(path): return path
     mgm, lfn = split_mgm(path, mgm=mgm)
+    lfn = osp.normpath(lfn)
     return _join_mgm_lfn(mgm, lfn)
 
+def dirname(path):
+    """
+    Like osp.dirname, but works with an mgm.
+    """
+    is_remote = has_protocol(path)
+    if is_remote:
+        mgm, path = split_mgm(path)
+    path = osp.dirname(osp.normpath(path))
+    return format(path, mgm) if is_remote else path
+
+def normpath(path):
+    """
+    Like osp.normpath, but works with an mgm.
+    """
+    is_remote = has_protocol(path)
+    if is_remote:
+        mgm, path = split_mgm(path)
+    path = osp.normpath(path)
+    return format(path, mgm) if is_remote else path
+
+def relpath(path, start):
+    """
+    Like osp.relpath, but works with an mgm.
+    """
+    if has_protocol(path) != has_protocol(start):
+        raise TypeError('{0} / {1}: either both or neither must have mgms'.format(path, start))
+    mgm1 = ''
+    mgm2 = ''
+    if has_protocol(path): mgm1, path = split_mgm(path)
+    if has_protocol(start): mgm2, start = split_mgm(start)
+    if mgm1 != mgm2:
+        raise TypeError('mgm mismatch: {0} vs. {1}'.format(mgm1, mgm2))
+    path = osp.normpath(path)
+    return osp.relpath(path, start)
+
+def iter_parent_dirs(path):
+    """
+    Iterates through all the parent directories of a path
+    E.g.:
+    `'/foo/bar'` --> `['/foo', '/']`
+    """
+    dir = dirname(path)
+    previous_dir = None
+    while dir != previous_dir:
+        yield dir
+        previous_dir = dir
+        dir = dirname(dir)
+
 def get_protocol(path, mgm=None):
     """
     Returns the protocol contained in the path string
@@ -303,14 +460,34 @@ class Inode(object):
         return stat(path)
 
     def __init__(self, path, modtime, isdir, size):
-        self.path = path
+        self.path = normpath(path)
         self.modtime = modtime
         self.isdir = isdir
         self.size = size
-        # Some derived properties
-        self.isfile = not(self.isdir)
-        self.size_human = bytes_to_human_readable(float(self.size))
-        self.basename = osp.basename(path)
+
+    @property
+    def isfile(self):
+        return not(self.isdir)
+
+    @property
+    def size_human(self):
+        return bytes_to_human_readable(float(self.size))
+
+    @property
+    def basename(self):
+        return osp.basename(self.path)
+
+    @property
+    def dirname(self):
+        return dirname(self.path)
+
+    @property
+    def path_no_mgm(self):
+        return split_mgm(self.path)[1]
+
+    @property
+    def mgm(self):
+        return split_mgm(self.path)[0]
 
     def __repr__(self):
         if len(self.path) > 40:
@@ -319,6 +496,41 @@ class Inode(object):
             shortpath = self.path
         return super(Inode, self).__repr__().replace('object', shortpath)
 
+    def __eq__(self, other):
+        return (self.path == other.path
+            and self.modtime == other.modtime
+            and self.isdir == other.isdir
+            and self.size == other.size
+            )
+
+class ExceptionWithPath(Exception):
+    """
+    Exception that optionally formats the error string with a path, if it is specified.
+    """
+    def __init__(self, msg, path=''):
+        super(Exception, self).__init__(msg + ((': ' + path) if path else ''))
+
+class NoSuchPath(ExceptionWithPath):
+    def __init__(self, path=''):
+        super(ExceptionWithPath, self).__init__('No such path', path)
+
+class PermissionDenied(ExceptionWithPath):
+    def __init__(self, path=''):
+        super(ExceptionWithPath, self).__init__('Permission denied', path)
+
+class HostUnreachable(ExceptionWithPath):
+    def __init__(self, path=''):
+        super(ExceptionWithPath, self).__init__('Host unreachable', path)
+
+class RmSafetyTrigger(ExceptionWithPath):
+    def __init__(self, path=''):
+        super(ExceptionWithPath, self).__init__('rm operation attempted on unsafe path', path)
+
+class NonZeroExitCode(subprocess.CalledProcessError):
+    def __init__(self, exitcode='?', cmd='?'):
+        super(Exception, self).__init__('Exit code {} for command {}'.format(cmd, exitcode))
+
+
 # _______________________________________________________
 # Cache
 
@@ -442,123 +654,143 @@ def load_tarball_cache(tarball, dst=None):
 # _______________________________________________________
 # Helpers for interactions with SE
 
-N_COPY_RETRIES = 0
 
-# Import implementations
-from . import xrd
-from . import pyxrd
-from . import gfal
-from . import eos
-from . import ssh
-_implementations = [ xrd, pyxrd, gfal, eos, ssh ]
-
-def get_implementation(name):
-    """
-    Takes the name of a submodule that contains implementations. Returns that module object.
-    """
-    for implementation in _implementations:
-        if implementation.__name__.rsplit('.',1)[-1] == name:
-            break
-    else:
-        raise Exception('No such implementation: {0}'.format(name))
-    return implementation
-
-_commands = [
+_valid_commands = [
     'mkdir', 'rm', 'stat', 'exists', 'isdir',
     'isfile', 'is_file_or_dir', 'listdir', 'cp', 'cat'
     ]
 
-def get_command(cmd_name, implementation=None, path=None):
-    """
-    Returns an implementation of a command.
-    Parameter `implementation` is a module object; it can also be a name of a module.
-    If it's None, a heuristic is ran to guess the best implementation.
-    The `path` parameter is optional and only serves to make a better guess for the implementation.
-    """
-    if not cmd_name in _commands:
-        raise Exception('Invalid command: {0}'.format(cmd_name))
-    if implementation is None:
-        implementation = best_implementation_heuristic(cmd_name, path)[0]
-    elif is_string(implementation):
-        implementation = get_implementation(implementation)
-    if not hasattr(implementation, cmd_name):
-        raise Exception(
-            'Implementation {0} has no function {1}'
-            .format(implementation.__name__, cmd_name)
-            )
-    return getattr(implementation, cmd_name)
+class Implementation:
 
-def best_implementation_heuristic(cmd_name, path=None):
+    rcodes = {}
+
+    def __init__(self):
+        self._is_installed = None
+
+    def run_command(self, *args, **kwargs):
+        """
+        Wrapper around `run_command` that inserts the return codes dict for proper
+        error handling
+        """
+        kwargs['rcodes'] = self.rcodes
+        return run_command(*args, **kwargs)
+
+    def is_installed(self):
+        if self._is_installed is None:
+            self._is_installed = self.check_is_installed()
+        return self._is_installed
+
+    def check_is_installed(self):
+        raise NotImplementedError
+
+    @add_env_kwarg
+    def exists(self, path):
+        try:
+            self.stat(path)
+            return True
+        except NoSuchPath:
+            return False
+
+    @add_env_kwarg
+    def isfile(self, path):
+        try:
+            inode = self.stat(path)
+            return inode.isfile
+        except NoSuchPath:
+            return False
+
+    @add_env_kwarg
+    def is_file_or_dir(self, path):
+        try:
+            inode = self.stat(path)
+            return 2 if inode.isfile else 1
+        except NoSuchPath:
+            return 0
+
+    @add_env_kwarg
+    def isdir(self, directory):
+        try:
+            inode = self.stat(directory)
+            return inode.isdir
+        except NoSuchPath:
+            return False
+
+
+from .gfal_implementation import GfalImplementation
+gfal = GfalImplementation()
+
+# from .pyxrd_implementation import PyxrdImplementation
+# pyxrd = PyxrdImplementation()
+
+from .xrd_implementation import XrdImplementation
+xrd = XrdImplementation()
+
+
+class PlaceholderImplementation(Implementation):
+    def check_is_installed(self):
+        return False
+
+pyxrd = PlaceholderImplementation()
+eos = PlaceholderImplementation()
+ssh = PlaceholderImplementation()
+
+
+implementations = dict(gfal=gfal, pyxrd=pyxrd, xrd=xrd, eos=eos, ssh=ssh)
+
+def get_implementation(implementation_name):
     """
-    Determines the best implementation for a path.
-    The order is loosely as follows:
-    - If the python bindings of xrootd are importable, those are probably the most reliable
-    - If the path does not start with root://, it's safer to use gfal
-    - For root://-starting paths, use xrootd
+    Returns an implementation instance corresponding to the passed name.
+    Returns None if `implementation_name` is 'auto' or None.
+    """
+    if implementation_name in ['auto', None]:
+        return None
+    return implementations[implementation_name]
+
+def best_implementation(cmd_name, path=None):
+    """
+    Given a command name, returns an installed implementation that has this command
     """
-    # Shortcut for ssh
     if path and is_ssh(path):
         logger.debug('Path is ssh-like')
-        return [ssh]
-    # Grid protocols
-    preferred_order = []
-    def check(module):
-        if module.is_installed() and hasattr(module, cmd_name) and not(module in preferred_order):
-            preferred_order.append(module)
-    # If this concerns removal, eos is the only tool that can do recursive directory removal:
-    if cmd_name == 'rm': check(eos)
-    # If path starts with root://, prefer xrd over gfal, otherwise the other way around
-    if not(path is None) and get_protocol(path) != 'root':
-        check(gfal)
-        check(pyxrd)
-        check(xrd)
+        preferred_order = [ssh]
+    elif cmd_name == 'rm':
+        preferred_order = [ eos, gfal, pyxrd, xrd]
     else:
-        check(pyxrd)
-        check(xrd)
-        check(gfal)
-    # Prefer eos last
-    check(eos)
-    # Check if at least one method is found
-    if not len(preferred_order):
-        raise Exception(
-            'No good implementation could be found for cmd {0} (path {1})'
-            .format(cmd_name, path)
-            )
-    logger.info('Using module %s to execute \'%s\' (path: %s)', preferred_order[0].__name__, cmd_name, path)
-    return preferred_order
+        preferred_order = [ xrd, gfal, pyxrd, eos ]
+    # Return first one that's installed
+    for implementation in preferred_order:
+        if implementation.is_installed() and hasattr(implementation, cmd_name):
+            logger.info(
+                'Using implementation %s to execute \'%s\' (path: %s)',
+                implementation.__class__.__name__, cmd_name, path
+                )
+            return implementation
+    raise Exception('No installed implementation found for cmd {0}, path {1}'.format(cmd_name, path))
 
-ENV = None
-def set_env(env):
-    """
-    Sets the env in which command line arguments are ran by default
-    """
-    global ENV
-    ENV = env
 
-@contextmanager
-def env_context(env):
+def make_global_scope_command(cmd_name):
     """
-    Temporarily sets an environment, and then reverts to the old environment
+    Creates a global scope command in case the user does not care about the
+    underlying implementation.
     """
-    global ENV
-    old_ENV = ENV
-    ENV = env
-    try:
-        yield None
-    finally:
-        ENV = old_ENV
-
-def add_env_kwarg(fn):
-    """
-    Function decorator that gives the function the `env` keyword argument
-    """
-    def wrapper(*args, **kwargs):
-        if 'env' in kwargs:
-            with env_context(kwargs.pop('env')):
-                return fn(*args, **kwargs)
-        else:
-            return fn(*args, **kwargs)
+    def wrapper(path, *args, **kwargs):
+        implementation = kwargs.pop('implementation', None)
+        if implementation is None: implementation = best_implementation(cmd_name, path)
+        return getattr(implementation, cmd_name)(path, *args, **kwargs)
     return wrapper
+        
+mkdir = make_global_scope_command('mkdir')
+rm = make_global_scope_command('rm')
+stat = make_global_scope_command('stat')
+exists = make_global_scope_command('exists')
+isdir = make_global_scope_command('isdir')
+isfile = make_global_scope_command('isfile')
+is_file_or_dir = make_global_scope_command('is_file_or_dir')
+listdir = make_global_scope_command('listdir')
+cat = make_global_scope_command('cat')
+cp = make_global_scope_command('cp')
+stat_fn = stat # Alias for if stat is a keyword in a function in this module
+
 
 # _______________________________________________________
 # Actual interactions with SE
@@ -567,141 +799,28 @@ def add_env_kwarg(fn):
 # to None, the 'best' implementation is guessed.
 
 @add_env_kwarg
-def mkdir(path, implementation=None):
+def put(path, contents='', make_parent_dirs=True, tmpfile_path='seutils_tmpfile', **cp_kwargs):
     """
-    Creates a directory on the SE
-    Does not check if directory already exists
+    Creates a file on a storage element.
+    `path` should contain an mgm
     """
-    path = format(path) # Ensures format
-    logger.warning('Creating directory on SE: {0}'.format(path))
-    cmd = get_command('mkdir', implementation=implementation, path=path)
-    cmd(path)
+    path = normpath(path)
+    tmpfile_path = osp.abspath(tmpfile_path)
+    if not has_protocol(path):
+        raise TypeError('Path {0} does not contain an mgm'.format(path))
+    # Open a local file
+    with open(tmpfile_path, 'w') as f:
+        f.write(contents)
+    try:
+        cp(tmpfile_path, path, **cp_kwargs)
+    finally:
+        os.remove(tmpfile_path)
 
-@add_env_kwarg
-def rm(path, recursive=False, implementation=None):
-    """
-    Creates a path on the SE
-    Does not check if path already exists
-    """
-    path = format(path) # Ensures format
-    logger.warning('Removing path on SE: {0}'.format(path))
-    cmd = get_command('rm', implementation=implementation, path=path)
-    cmd(path, recursive=recursive)
-
-@add_env_kwarg
-@cache
-def stat(path, not_exist_ok=False, implementation=None):
-    """
-    Returns an Inode object for path.
-    If not_exist_ok is True and the path doesn't exist, it returns None
-    without raising an exception
-    """
-    path = format(path) # Ensures format
-    cmd = get_command('stat', implementation=implementation, path=path)
-    return cmd(path, not_exist_ok=not_exist_ok)
-
-@add_env_kwarg
-def stat_function(*args, **kwargs):
-    """
-    Alternative name for the stat function, since stat is also an often used keyword in functions
-    """
-    return stat(*args, **kwargs)
-
-@add_env_kwarg
-@cache
-def exists(path, implementation=None):
-    """
-    Returns a boolean indicating whether the path exists.
-    """
-    return get_command('exists', implementation=implementation, path=path)(path)
-
-@add_env_kwarg
-@cache
-def isdir(path, implementation=None):
-    """
-    Returns a boolean indicating whether the directory exists.
-    Also returns False if the passed path is a file.
-    """
-    return get_command('isdir', implementation=implementation, path=path)(path)
-
-@add_env_kwarg
-@cache
-def isfile(path, implementation=None):
-    """
-    Returns a boolean indicating whether the file exists.
-    Also returns False if the passed path is a directory.
-    """
-    return get_command('isfile', implementation=implementation, path=path)(path)
-
-@add_env_kwarg
-@cache
-def is_file_or_dir(path, implementation=None):
-    """
-    Returns 0 if path does not exist
-    Returns 1 if it's a directory
-    Returns 2 if it's a file
-    """
-    return get_command('is_file_or_dir', implementation=implementation, path=path)(path)
-
-@add_env_kwarg
-def listdir(path, stat=False, assume_directory=False, implementation=None):
-    """
-    Returns the contents of a directory
-    If 'assume_directory' is True, it is assumed the user took
-    care to pass a path to a valid directory, and no check is performed
-    """
-    if not(assume_directory) and not isdir(path):
-        raise Exception('{0} is not a directory'.format(path))
-    cmd = get_command('listdir', implementation=implementation, path=path)
-    return cmd(path, stat)
-
-# Custom cache key generator dependent on whether stat was True or False
-def _listdir_keygen(path, stat=False, *args, **kwargs):
-    return path.strip() + '___stat{}'.format(stat)
-listdir.keygen = _listdir_keygen
-listdir = cache(listdir)
-
-@add_env_kwarg
-def cp(src, dst, implementation=None, **kwargs):
-    """
-    Copies a file `src` to the storage element.
-    Does not format `src` or `dst`; user is responsible for formatting.
-
-    The method can be 'auto', 'xrdcp', or 'gfal-copy'. If 'auto', a heuristic
-    will be applied to determine whether to best use xrdcp or gfal-copy.
-    """
-    logger.warning('Copying %s --> %s', src, dst)    
-    cmd = get_command('cp', implementation=implementation, path=dst if has_protocol(dst) else src)
-    cmd(src, dst, **kwargs)
-
-@add_env_kwarg
-def cat(path, implementation=None):
-    """
-    Returns the contents of a directory
-    If 'assume_directory' is True, it is assumed the user took
-    care to pass a path to a valid directory, and no check is performed
-    """
-    cmd = get_command('cat', implementation=implementation, path=path)
-    return cmd(path)
-
-# _______________________________________________________
-# Algorithms that use the SE interaction implementation, but are agnostic to the underlying tool
-
-def cp_to_se(src, dst, **kwargs):
-    """
-    Like cp, but assumes dst is a location on a storage element and src is local
-    """
-    cp(src, format(dst), **kwargs)
-
-def cp_from_se(src, dst, **kwargs):
-    """
-    Like cp, but assumes src is a location on a storage element and dst is local
-    """
-    cp(format(src), dst, **kwargs)
 
 MAX_RECURSION_DEPTH = 20
 
-def ls(path, stat=False, assume_directory=False, no_expand_directory=False):
+@add_env_kwarg
+def ls(path, stat=False, assume_isdir=False, no_expand_directory=False, implementation=None):
     """
     Lists all files and directories in a directory on the SE.
     It first checks whether the path exists and is a file or a directory.
@@ -711,32 +830,32 @@ def ls(path, stat=False, assume_directory=False, no_expand_directory=False):
 
     If stat is True, it returns Inode objects which contain more information beyond just the path
 
-    If assume_directory is True, the first check is not performed and the algorithm assumes
+    If assume_isdir is True, the first check is not performed and the algorithm assumes
     the user took care to pass a path to a directory. This saves a request to the SE, which might
-    matter in the walk algorithm. For singular use, assume_directory should be set to False.
+    matter in the walk algorithm. For singular use, assume_isdir should be set to False.
 
     If no_expand_directory is True, the contents of the directory are not listed, and instead
     a formatted path to the directory is returned (similar to unix's ls -d)
     """
     path = format(path)
-    if assume_directory:
+    if assume_isdir:
         status = 1
     else:
-        status = is_file_or_dir(path)
+        status = is_file_or_dir(path, implementation=implementation)
     # Depending on status, return formatted path to file, directory contents, or raise
     if status == 0:
-        raise RuntimeError('Path \'{0}\' does not exist'.format(path))
+        raise NoSuchPath(path)
     elif status == 1:
         # It's a directory
         if no_expand_directory:
             # If not expanding, just return a formatted path to the directory
-            return [stat_function(path) if stat else path]
+            return [stat_fn(path, implementation=implementation) if stat else path]
         else:
             # List the contents of the directory
-            return listdir(path, assume_directory=True, stat=stat) # No need to re-check whether it's a directory
+            return listdir(path, assume_isdir=True, stat=stat, implementation=implementation) # No need to re-check whether it's a directory
     elif status == 2:
         # It's a file; just return the path to the file
-        return [stat_function(path) if stat else path]
+        return [stat_fn(path, implementation=implementation) if stat else path]
 
 class Counter:
     """
@@ -748,7 +867,8 @@ class Counter:
     def plus_one(self):
         self.i += 1
 
-def walk(path, stat=False):
+@add_env_kwarg
+def walk(path, stat=False, implementation=None):
     """
     Entry point for walk algorithm.
     Performs a check whether the starting path is a directory,
@@ -757,17 +877,17 @@ def walk(path, stat=False):
     made to the storage element, so that 'accidents' are limited
     """
     path = format(path)
-    status = is_file_or_dir(path)
+    status = is_file_or_dir(path, implementation=implementation)
     if not status == 1:
         raise RuntimeError(
             '{0} is not a directory'
             .format(path)
             )
     counter = Counter()
-    for i in _walk(path, stat, counter):
+    for i in _walk(path, stat, counter, implementation=implementation):
         yield i
 
-def _walk(path, stat, counter):
+def _walk(path, stat, counter, implementation=None):
     """
     Recursively calls ls on traversed directories.
     The yielded directories list can be modified in place
@@ -780,7 +900,7 @@ def _walk(path, stat, counter):
             ' set seutils.MAX_RECURSION_DEPTH to a larger number.'
             .format(MAX_RECURSION_DEPTH)
             )
-    contents = ls(path, stat=True, assume_directory=True)
+    contents = ls(path, stat=True, assume_isdir=True, implementation=implementation)
     counter.plus_one()
     files = [ c for c in contents if c.isfile ]
     files.sort(key=lambda f: f.basename)
@@ -795,51 +915,11 @@ def _walk(path, stat, counter):
         # dirnames after yield
         directories = [ d for d in directories if d.path in dirnames ]
     for directory in directories:
-        for i in _walk(directory.path, stat, counter):
+        for i in _walk(directory.path, stat, counter, implementation=implementation):
             yield i
 
-def ls_root(paths):
-    """
-    Flexible function that attempts to return a list of root files based on what
-    the user most likely wanted to query.
-    Takes a list of paths as input. If input as a string, it will be turned into a len-1 list.
-    Firstly it is checked whether the path exists locally.
-      If it's a root file, it's appended to the output,
-      If it's a directory, it will be globbed for *.root.
-    Secondly it's attempted to reach the path remotely.
-    Returns a list of .root files.
-    """
-    if is_string(paths): paths = [paths]
-    root_files = []
-    for path in paths:
-        if osp.exists(path):
-            # Treat as a local path
-            if osp.isfile(path):
-                if path.endswith('.root'):
-                    root_files.append(path)
-            elif osp.isdir(path):
-                root_files.extend(glob.glob(osp.join(path, '*.root')))
-        else:
-            # Treat path as a SE path
-            try:
-                stat = is_file_or_dir(path)
-                if stat == 1:
-                    # It's a directory
-                    root_files.extend([ f for f in ls(path) if f.endswith('.root') ])
-                elif stat == 2:
-                    # It's a file
-                    root_files.append(format(path))
-                elif stat == 0:
-                    logger.warning('Path %s does not exist locally or remotely', path)
-            except RuntimeError:
-                logger.warning(
-                    'Path %s does not exist locally and could not be treated as a remote path',
-                    path
-                    )
-    root_files.sort()
-    return root_files
-
-def ls_wildcard(pattern, stat=False):
+@add_env_kwarg
+def ls_wildcard(pattern, stat=False, implementation=None):
     """
     Like ls, but accepts wildcards * .
     Directories are *not* expanded.
@@ -850,7 +930,7 @@ def ls_wildcard(pattern, stat=False):
     """
     pattern = format(pattern)
     if not '*' in pattern:
-        return ls(pattern, stat=stat, no_expand_directory=True)
+        return ls(pattern, stat=stat, no_expand_directory=True, implementation=implementation)
     import re
     if not stat and not '*' in pattern.rsplit('/',1)[0]:
         # If there is no star in any part but the last one and we don't need to stat, it is
@@ -859,7 +939,7 @@ def ls_wildcard(pattern, stat=False):
         # only for the very last part'
         logger.info('Detected * only in very last part of pattern and stat=False; using shortcut')
         directory, pattern = pattern.rsplit('/',1)
-        contents = ls(directory)
+        contents = ls(directory, implementation=implementation)
         if pattern == '*':
             # Skip the regex matching if set to 'match all'
             return contents
@@ -873,7 +953,7 @@ def ls_wildcard(pattern, stat=False):
     base = pattern.split('*',1)[0].rsplit('/',1)[0]
     logger.debug('Found base pattern %s from pattern %s', base, pattern)
     matches = []
-    for path, directories, files in walk(base, stat=stat):
+    for path, directories, files in walk(base, stat=stat, implementation=implementation):
         level = path.count('/')
         logger.debug('Level is %s for path %s', level, path)
         trimmed_pattern = '/'.join(pattern.split('/')[:level+2]).replace('*', '.*')
@@ -894,77 +974,77 @@ def ls_wildcard(pattern, stat=False):
             directories[:] = []
     return matches
 
-# _______________________________________________________
-# Command line helpers
 
-class Parser:
+def listdir_recursive(directory, stat=False, implementation=None):
     """
-    Very thin wrapper class for argparse.ArgumentParser with some options
-    used for every command line tool in seutils
+    Returns a list of all paths (or Inodes if `stat=True`) recursively under `directory`.
     """
-    def __init__(self, *args, **kwargs):
-        self.parser = argparse.ArgumentParser(*args, **kwargs)
-        self.parser.add_argument('-v', '--verbose', action='store_true', help='Increases verbosity')
-        self.parser.add_argument('-d', '--dry', action='store_true', help='Does not actually run any commands')
+    contents = []
+    for path, directories, files in walk(directory, stat=stat, implementation=implementation):
+        contents.extend(directories)
+        contents.extend(files)
+    return contents
 
-    def add_argument(self, *args, **kwargs):
-        self.parser.add_argument(*args, **kwargs)
 
-    def parse_args(self, *args, **kwargs):
-        parsed_args = self.parser.parse_args(*args, **kwargs)
-        if parsed_args.verbose: debug()
-        if parsed_args.dry: drymode()
-        return parsed_args
-
-class ParserMultipleLFNs(Parser):
+def _sorted_paths_from_set(relpaths_set, relpaths, contents):
     """
-    Like ArgumentParser, but includes a multiple LFNs argument
+    Used internally by `diff`. For every element in `relpaths_set`,
+    return the matching item from `contents`. Preserves order of `contents`.
     """
-    def __init__(self, *args, **kwargs):
-        super(ParserMultipleLFNs, self).__init__(self, *args, **kwargs)
-        self.parser.add_argument(
-            'lfns', type=str, nargs='+',
-            help='Paths to datasets (directories that contain .root files)'
-            )
+    selected_contents = []
+    for rpath, content in zip(relpaths, contents):
+        if rpath in relpaths_set:
+            selected_contents.append(content)
+    return selected_contents
 
-    def parse_args(self, *args, **kwargs):
-        parsed_args = super(ParserMultipleLFNs, self).parse_args(*args, **kwargs)
-        parsed_args.lfns = cli_expand_lfns(parsed_args.lfns)
-        return parsed_args
+@add_env_kwarg
+def diff(left, right, stat=False, implementation=None):
+    """
+    Returns 4 lists of paths (or Inodes, if `stat=True`):
+    - Paths in `left` that are also in `right` (intersection_left)
+    - Paths in `right` that are also in `left` (intersection_right)
+    - Paths in `left` that are not in `right` (only_in_left)
+    - Paths in `right` that are not in `left` (only_in_right)
 
+    (Note intersection_left and intersection_right) will have the same
+    contents, but different mgms)
 
-def cli_detect_fnal():
-    if DEFAULT_MGM is None and os.uname()[1].endswith('.fnal.gov'):
-        mgm = 'root://cmseos.fnal.gov'
-        logger.warning('Detected fnal.gov host; using mgm %s as default if necessary', mgm)
-        set_default_mgm(mgm)
+    TODO: Currently only implemented if both left and right are remote!
+    """
+    for path in [left, right]:
+        if not has_protocol(path):
+            raise NotImplementedError('diff does not support local paths yet: {0}'.format(path))
 
-def cli_flexible_format(lfn, mgm=None):
-    if is_ssh(lfn): return lfn
-    cli_detect_fnal()
-    if not has_protocol(lfn) and not lfn.startswith('/'):
-        try:
-            prefix = '/store/user/' + os.environ['USER']
-            logger.warning('Pre-fixing %s', prefix)
-            lfn = os.path.join(prefix, lfn)
-        except KeyError:
-            pass
-    if has_protocol(lfn):
-        return format(lfn)
+    contents_left = listdir_recursive(left)
+    contents_right = listdir_recursive(right)
+
+    if stat:
+        paths_left = [ n.path for n in contents_left ]
+        paths_right = [ n.path for n in contents_right ]
     else:
-        return format(lfn, mgm)
+        paths_left = contents_left
+        paths_right = contents_right
+        
+    relpaths_left = [ relpath(p, left) for p in paths_left ]
+    relpaths_right = [ relpath(p, right) for p in paths_right ]
+    set_relpaths_left = set(relpaths_left)
+    set_relpaths_right = set(relpaths_right)
 
-def cli_expand_lfns(raw_lfns):
-    # First preprocess any wildcards
-    lfns = []
-    for lfn in raw_lfns:
-        if '*' in lfn:
-            lfns.extend(ls_wildcard(lfn))
-        else:
-            lfns.append(lfn)
-    # Format once more to be sure everything is a proper lfn
-    lfns = [cli_flexible_format(lfn, cli_detect_fnal()) for lfn in lfns]
-    return lfns
+    intersection = set_relpaths_left.intersection(set_relpaths_right)
+    only_in_left = set_relpaths_left - set_relpaths_right
+    only_in_right = set_relpaths_right - set_relpaths_left
+
+    return (
+        _sorted_paths_from_set(intersection, relpaths_left, contents_left),
+        _sorted_paths_from_set(intersection, relpaths_right, contents_right),
+        _sorted_paths_from_set(only_in_left, relpaths_left, contents_left),
+        _sorted_paths_from_set(only_in_right, relpaths_right, contents_right),
+        )
+
+# _______________________________________________________
+# CLI
+
+from . import cli
 
 # _______________________________________________________
 # root utils extension
